import datetime
import ssl
from typing import Optional

import httpx
from h2o_authn import TokenProvider, token
from h2o_authn.error import TokenEndpointError
from h2o_authn.provider import (DEFAULT_EXPIRES_IN_FALLBACK,
                                DEFAULT_EXPIRY_THRESHOLD, DEFAULT_HTTP_TIMEOUT)

DEFAULT_TOKEN_FAIL_RETRY_MAX_COUNT: int = 3


class ExtendedTokenProvider(TokenProvider):

    def __init__(
            self,
            *,
            client_id: str,
            username: str = None,
            password: str = None,
            issuer_url: Optional[str] = None,
            token_endpoint_url: Optional[str] = None,
            client_secret: Optional[str] = None,
            scope: Optional[str] = None,
            expiry_threshold: datetime.timedelta = DEFAULT_EXPIRY_THRESHOLD,
            expires_in_fallback: datetime.timedelta = DEFAULT_EXPIRES_IN_FALLBACK,
            http_timeout: datetime.timedelta = DEFAULT_HTTP_TIMEOUT,
            minimal_refresh_period: Optional[datetime.timedelta] = None,
            http_ssl_context: Optional[ssl.SSLContext] = None,
            token_generation_retry_max_count: Optional[
                int
            ] = DEFAULT_TOKEN_FAIL_RETRY_MAX_COUNT,
    ) -> None:
        self.token_generation_retry_max_count = token_generation_retry_max_count
        self.token_generation_retry_count = token_generation_retry_max_count

        self._username = username
        self._password = password

        # Need to initialize the parent first, to call
        # _ensure_token_endpoint_url() method
        self.init_parent(
            client_id,
            client_secret,
            expires_in_fallback,
            expiry_threshold,
            http_ssl_context,
            http_timeout,
            issuer_url,
            minimal_refresh_period,
            scope,
            token_endpoint_url,
        )
        self._ensure_token_endpoint_url()

        # Update parent class with the refresh token
        refresh_token = self._fetch_refresh_token_with_password()
        self.init_parent(
            client_id,
            client_secret,
            expires_in_fallback,
            expiry_threshold,
            http_ssl_context,
            http_timeout,
            issuer_url,
            minimal_refresh_period,
            scope,
            token_endpoint_url,
            refresh_token,
        )

    def init_parent(
            self,
            client_id,
            client_secret,
            expires_in_fallback,
            expiry_threshold,
            http_ssl_context,
            http_timeout,
            issuer_url,
            minimal_refresh_period,
            scope,
            token_endpoint_url,
            refresh_token="",
    ):
        super().__init__(
            refresh_token=refresh_token,
            client_id=client_id,
            issuer_url=issuer_url,
            token_endpoint_url=token_endpoint_url,
            client_secret=client_secret,
            scope=scope,
            expiry_threshold=expiry_threshold,
            expires_in_fallback=expires_in_fallback,
            http_timeout=http_timeout,
            minimal_refresh_period=minimal_refresh_period,
            http_ssl_context=http_ssl_context,
        )

    def _call_token_endpoint_with_password(self):
        with self._client() as client:
            resp = client.post(
                self._token_endpoint_url,
                data=self._create_password_grant_request_data(),
            )
        return resp

    def _fetch_refresh_token_with_password(self):
        resp = self._call_token_endpoint_with_password()
        resp_data = resp.json()

        try:
            resp.raise_for_status()
        except httpx.HTTPStatusError:
            if resp.status_code != 400 or not resp_data.get("error"):
                raise
            raise TokenEndpointError(
                error=resp_data["error"],
                error_description=resp_data.get("error_description"),
                error_uri=resp_data.get("error_uri"),
            ) from None

        return resp_data.get("refresh_token")

    def _create_password_grant_request_data(self):
        data = {
            "grant_type": "password",
            "client_id": self._client_id,
            "username": self._username,
            "password": self._password,
        }

        if self._client_secret:
            data["client_secret"] = self._client_secret

        if self._scope:
            data["scope"] = self._scope

        return data

    def token(self) -> token.Token:
        try:
            token = super().token()
            self.token_generation_retry_count = self.token_generation_retry_max_count
            return token
        except Exception as e:
            print(f"Token generation failed with the refresh grant!, {e}")
            self.token_generation_retry_count -= 1
            if self.token_generation_retry_count >= 0:
                print(f"Retrying with the password grant type...")
                resp = self._call_token_endpoint_with_password()
                self.token_generation_retry_count = (
                    self.token_generation_retry_max_count
                )
                self._update_token(resp)



import json
import os

import h2o_authn
import requests
from h2ogpte import H2OGPTE

from .auth import ExtendedTokenProvider


class RequestError(Exception):
    def __init__(self, message):
        self.message = message


class RemoteError(Exception):
    def __init__(self, message):
        self.message = message


def get_token_endpoint(issuer_url):
    # Call OIDC protocol standard url to retrieve token endpoint
    uri = issuer_url.rstrip("/") + "/.well-known/openid-configuration"

    res = requests.get(url=uri)
    try:
        res.raise_for_status()
    except requests.HTTPError as e:
        msg = f"Authentication server responded with {res.status_code}."
        print(f"[ERROR] {msg}\n\n{res.content}")
        raise RequestError(msg) from e

    try:
        response_json = res.json()
    except json.JSONDecodeError as e:
        msg = "Authentication server response is not a valid JSON."
        print(f"[ERROR] {msg}\n\n{res.content}")
        raise RequestError(msg) from e

    if "error" in response_json:
        raise RemoteError(response_json["error"])

    if "token_endpoint" in response_json:
        return response_json["token_endpoint"]
    else:
        raise Exception(
            f'Authentication endpoint discovery failed. {response_json["error"]}'
        )


def connect_to_h2ogpte(refresh_token):
    # Precedence given in the following order;
    # 1. If H2OGPTE_API_TOKEN is set, use it without a token provider
    # 2. If OIDC_AUTH_URL is set, use it to generate a refresh token
    # 3. If H2O_WAVE_OIDC_PROVIDER_URL is set, use it along with q.auth.refresh_token
    # 4. Or else raise an error
    gpte_api_token = os.getenv("H2OGPTE_API_TOKEN")

    if gpte_api_token:
        print(
            f"[H2O_GPTE_CLIENT] H2OGPTE_API_TOKEN found, trying to "
            f"connect to {os.getenv('H2OGPTE_URL')} using "
            f"H2OGPTE_API_TOKEN: {gpte_api_token[:5]}..."
            f"{gpte_api_token[-5:]}"
        )

        return H2OGPTE(
            address=os.getenv("H2OGPTE_URL"),
            api_key=os.getenv("H2OGPTE_API_TOKEN"),
        )

    else:
        if os.getenv("OIDC_AUTH_URL"):
            print(
                f"[H2O_GPTE_CLIENT] OIDC_AUTH_URL found, trying to "
                f"connect to {os.getenv('OIDC_AUTH_URL')} to generate "
                f"a refresh token..."
            )
            auth_url = os.getenv("OIDC_AUTH_URL")
            token_endpoint = get_token_endpoint(auth_url)

            client_id = os.getenv("OIDC_CLIENT_ID")
            client_secret = os.getenv("OIDC_CLIENT_SECRET")
            username = os.getenv("OIDC_USERNAME")
            password = os.getenv("OIDC_PASSWORD")

            client_secret_display = (
                f"{client_secret[:3]}...{client_secret[-3:]}"
                if client_secret is not None
                else ""
            )
            print(
                f"[H2O_GPTE_CLIENT] Generating a refresh_token with "
                f"token_endpoint: {token_endpoint} "
                f"client_id: {client_id} "
                f"{client_secret_display} "
                f"username: {username} "
                f"with password: {'*' * len(password)}..."
            )
            token_provider = ExtendedTokenProvider(
                username=username,
                password=password,
                token_endpoint_url=token_endpoint,
                client_id=client_id,
                client_secret=client_secret if client_secret else None,
            )
        elif refresh_token is not None:
            print(
                f"[H2O_GPTE_CLIENT] q.auth.refresh_token found, trying to "
                f"connect to {os.getenv('OIDC_PROVIDER_URL')}..."
            )
            token_endpoint = get_token_endpoint(os.getenv("OIDC_PROVIDER_URL"))
            client_id = os.getenv("OIDC_CLIENT_ID")
            client_secret = os.getenv("OIDC_CLIENT_SECRET")

            client_secret_display = (
                f"{client_secret[:3]}...{client_secret[-3:]}"
                if client_secret is not None
                else ""
            )
            print(
                f"[H2O_GPTE_CLIENT] Using token provider with "
                f"token_endpoint: {token_endpoint} "
                f"client_id: {client_id} "
                f"{client_secret_display} "
                f"with refresh_token: {refresh_token[:3]}...{refresh_token[-3:]}..."
            )
            token_provider = h2o_authn.TokenProvider(
                refresh_token=refresh_token,
                token_endpoint_url=token_endpoint,
                client_id=client_id,
                client_secret=client_secret if client_secret else None,
            )
        else:
            raise Exception(
                "You need to set one of H2OGPTE_API_TOKEN, "
                "OIDC_AUTH_URL or H2O_WAVE_OIDC_PROVIDER_URL"
            )
        return H2OGPTE(address=os.getenv("H2OGPTE_URL"), token_provider=token_provider)
